/*
    Copyright (C) 2000 Paul Barton-Davis 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: ardour_ui_ed.cc,v 1.1 2000/11/03 19:45:39 pbd Exp $
*/

/* This file contains any ARDOUR_UI methods that require knowledge of
   the editor, and exists so that no compilation dependency exists
   between the main ARDOUR_UI modules and the Editor class. This
   is to cut down on the nasty compile times for both these classes.
*/

#include "ardour_ui.h"
#include "editor.h"
#include "tape.h"
#include "track_ui.h"
#include "audio_clock_ui.h"
#include "alsa_device.h"
#include "location_ui.h"

using namespace ARDOUR;

int
ARDOUR_UI::create_editor ()

{
	editor = new Editor ();

	if (!editor->ok()) {
		return -1;
	}

	notebook.pages().push_back (Gtk::Notebook_Helpers::TabElem (*editor, "Editor"));

	return 0;
}

void
ARDOUR_UI::switched_to_recorder ()

{
	start_metering ();
	if (editor) {
		editor->inactive ();
	}
}

void
ARDOUR_UI::switched_to_editor ()

{
	stop_metering ();
	if (editor) {
		editor->active ();
	}
}


void
ARDOUR_UI::connect_to_tape (Tape *t)

{
	tape = t;

	tape->TransportStateChange.connect 
		(slot (*this, &ARDOUR_UI::transport_change));
	tape->RecordEnabled.connect 
		(slot (*this, &ARDOUR_UI::record_status_change));
	tape->RecordDisabled.connect 
		(slot (*this, &ARDOUR_UI::record_status_change));
	tape->ControlChanged.connect
		(slot (*this, &ARDOUR_UI::map_control_change));

	map_tape_description ();
	xfade_button->set_sensitive (true);
	tape->set_xfade_length ((size_t) xfade_adjustment.get_value());
	clock_ui->connect_to_audiothing (tape);
	mcd->connect_to_audiothing (tape);
	location_ui->connect_to_tape (tape);

	for (size_t n = 0; n < Config->max_tracks; n++) {
		track_ui[n]->connect_to_tape (tape);
	}
	
	/* Timecode clock is on by default after the clock is
	   connected to a tape, so show that here.
	*/
	
	clock_button->set_active (true);
	map_button_state ();
	rec_enable_blink_on = true;
	blink_handler_tag = gtk_timeout_add (250, _blink_handler, this);

	editor->connect_to_tape (tape);
}


void
ARDOUR_UI::save_state ()

{
	/* XXX lots more to do here */

	if (tape) {
		tape->save_locations ();
		editor->save_all_tracks ();
	}
}

void
ARDOUR_UI::restore_state ()

{
	/* XXX lots more to do here */

	if (tape) {
		tape->restore_locations ();
	}
}
