/*
    Copyright (C) 2000 Paul Barton-Davis 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: location_ui.cc,v 1.26 2000/10/29 04:42:51 pbd Exp $
*/

#include <math.h>
#include <gtkmmext/utils.h>
#include <gtkmmext/led.h>
#include <gtkmmext/pixmap_button.h>

#include "location_ui.h"
#include "alsa_device.h"
#include "utils.h"
#include "configuration.h"
#include "tape.h"
#include "ardour_ui.h"

using namespace ARDOUR;
using namespace Gtkmmext;

Location_UI::Location_UI (ALSA::MultiChannelDevice *mcdp, 
			  Gtk::Widget &pixmap_window)

	: mcd (mcdp),
	  
	  start_adjustment (0, 0, UINT_MAX, 48000, 48000 * 60),
	  start_spinner (start_adjustment, 48000 * 30),

	  end_adjustment (0, 0, UINT_MAX, 48000, 48000 * 60),
	  end_spinner (end_adjustment, 48000 * 30),

	  start_grab_tape (ARDOUR_UI::instance()->small_push_button_pix(),
			   "LocationSetButton", false),
	  start_forward (ARDOUR_UI::instance()->small_push_button_pix(),
			   "LocationSetButton", false),
	  start_backward (ARDOUR_UI::instance()->small_push_button_pix(),
			   "LocationSetButton", false),
	  start_forward_label ("FWD"),
	  start_grab_label ("SET"),
	  start_backward_label ("BACK"),

	  end_grab_tape (ARDOUR_UI::instance()->small_push_button_pix(),
			 "LocationSetButton", false),
	  end_forward (ARDOUR_UI::instance()->small_push_button_pix(),
			 "LocationSetButton", false),
	  end_backward(ARDOUR_UI::instance()->small_push_button_pix(),
			 "LocationSetButton", false),
	  end_forward_label ("FWD"),
	  end_grab_label ("SET"),
	  end_backward_label ("BACK"),

	  next_location (ARDOUR_UI::instance()->small_push_button_pix(),
			   "LocationSelectButton", false),
	  previous_location (ARDOUR_UI::instance()->small_push_button_pix(),
			 "LocationSelectButton", false),
	  next_label ("NEXT"),
	  previous_label ("PREVIOUS"),

	  start_label ("START"),
	  end_label ("END"),
	  auto_punch_label ("AUTO\nPUNCH?"),
	  auto_loop_label ("AUTO\nLOOP?"),

	  yes_its_me (false)
	
{
	gchar *pixmap_path;

	tape = 0;

	start_label.set_name ("LocationLabel");
	end_label.set_name ("LocationLabel");

	start_forward_label.set_name ("GlobalButtonLabel");
	start_forward_box.pack_start (start_forward_label, false, false);
	start_forward_box.pack_start (start_forward, false, false);

	start_grab_label.set_name ("GlobalButtonLabel");
	start_grab_box.pack_start (start_grab_label, false, false);
	start_grab_box.pack_start (start_grab_tape, false, false);

	start_backward_label.set_name ("GlobalButtonLabel");
	start_backward_box.pack_start (start_backward_label, false, false);
	start_backward_box.pack_start (start_backward, false, false);

	start_sub_hbox.set_spacing (5);
	start_sub_hbox.pack_start (start_backward_box, true, false);
	start_sub_hbox.pack_start (start_grab_box, true, false);
	start_sub_hbox.pack_start (start_forward_box, true, false);

	start_display.set_editable (false);
	start_display.unset_flags (GTK_CAN_FOCUS);
	start_display.set_name ("LocationDisplay");
	start_frame.add (start_display);
	start_frame.set_shadow_type (GTK_SHADOW_IN);
	start_frame.set_name ("BaseFrame");
	start_vbox.pack_start (start_label, true, false, 5);
	start_vbox.pack_start (start_frame, false, false);
	start_vbox.pack_start (start_sub_hbox, false, false);

	end_forward_label.set_name ("GlobalButtonLabel");
	end_forward_box.pack_start (end_forward_label, false, false);
	end_forward_box.pack_start (end_forward, false, false);

	end_grab_label.set_name ("GlobalButtonLabel");
	end_grab_box.pack_start (end_grab_label, false, false);
	end_grab_box.pack_start (end_grab_tape, false, false);

	end_backward_label.set_name ("GlobalButtonLabel");
	end_backward_box.pack_start (end_backward_label, false, false);
	end_backward_box.pack_start (end_backward, false, false);

	end_sub_hbox.set_spacing (5);
	end_sub_hbox.pack_start (end_backward_box, true, false);
	end_sub_hbox.pack_start (end_grab_box, true, false);
	end_sub_hbox.pack_start (end_forward_box, true, false);

	end_display.set_editable (false);
	end_display.unset_flags (GTK_CAN_FOCUS);
	end_display.set_name ("LocationDisplay");
	end_frame.add (end_display);
	end_frame.set_shadow_type (GTK_SHADOW_IN);
	end_frame.set_name ("BaseFrame");
	end_vbox.pack_start (end_label, true, false, 5);
	end_vbox.pack_start (end_frame, false, false);
	end_vbox.pack_start (end_sub_hbox, false, false);

	name_entry.set_name ("LocationDisplay");
	name_frame.add (name_entry);
	name_frame.set_shadow_type (GTK_SHADOW_IN);
	name_frame.set_name ("BaseFrame");

	name_label.set_text ("DESCRIPTION");
	name_label.set_name ("LocationLabel");

	name_vbox.pack_start (name_label, false, false, 5);
	name_vbox.pack_start (name_frame, false, false);


	next_label.set_name ("LocationLabel");
	previous_label.set_name ("LocationLabel");

	previous_box.pack_start (previous_location, false, false);
	previous_box.pack_start (previous_label, false, false, 5);

	next_box.pack_start (next_label, false, false, 5);
	next_box.pack_start (next_location, false, false);

	name_vbox.pack_start (next_prev_hbox, false, false, 5);

	start_end_hbox.pack_start (name_vbox, true, true, 5);
	start_end_hbox.pack_start (start_vbox, false, false, 10);
	start_end_hbox.pack_start (end_vbox, false, false, 10);

	next_prev_hbox.pack_start (previous_box, true, false);
	next_prev_hbox.pack_start (next_box, true, false);

	left_vpacker.pack_start (start_end_hbox, false, false, 10);

	auto_punch_indicator = new PixmapButton 
		(ARDOUR_UI::instance()->toggle_button_pix(), 
		 "GlobalButtonLabel", false);

	auto_punch_label.set_name ("GlobalButtonLabel");
	auto_punch_box.set_spacing (5);
	auto_punch_box.pack_start (*auto_punch_indicator, false, false);
	auto_punch_box.pack_start (auto_punch_label, false, false);

	auto_loop_indicator = new PixmapButton 
		(ARDOUR_UI::instance()->toggle_button_pix(), 
		 "GlobalButtonLabel", false);

	auto_loop_label.set_name ("GlobalButtonLabel");
	auto_loop_box.set_spacing (5);
	auto_loop_box.pack_start (*auto_loop_indicator, false, false);
	auto_loop_box.pack_start (auto_loop_label, false, false);

	pixmap_path = gtk_rc_find_pixmap_in_path (0, Config->start_button_image.c_str());
	start_pixmap = new Gtk::Pixmap (pixmap_path);
	start_button.add (*start_pixmap);
	start_button_vbox.pack_start (start_button, false, false);

	pixmap_path = gtk_rc_find_pixmap_in_path (0, Config->end_button_image.c_str());
	end_pixmap = new Gtk::Pixmap (pixmap_path);
	end_button.add (*end_pixmap);
	end_button_vbox.pack_start (end_button, false, false);

	pixmap_path = gtk_rc_find_pixmap_in_path (0, Config->set_next_button_image.c_str());
	set_next_pixmap = new Gtk::Pixmap (pixmap_path);
	set_next_button.add (*set_next_pixmap);
	set_next_button_vbox.pack_start (set_next_button, false, false);

	auto_box.set_spacing (10);
	auto_box.pack_start (auto_punch_box, true, false);
	auto_box.pack_start (auto_loop_box, true, false);
	
	middle_vpacker.pack_start (start_button_vbox, true, false);
	middle_vpacker.pack_start (end_button_vbox, true, false, 10);

	right_vpacker.pack_start (set_next_button_vbox, true, false);

	hpacker.pack_start (left_vpacker, false, false);
	hpacker.pack_start (middle_vpacker, true, false, 10);
	hpacker.pack_start (right_vpacker, true, false, 10);
	hpacker.pack_start (auto_box, false, false, 10);

	name_entry.add_events (GDK_BUTTON_RELEASE_MASK);
	name_entry.changed.connect(slot(this,&Location_UI::set_location_name));
	
	next_location.clicked.connect (slot (this, &Location_UI::up));
	previous_location.clicked.connect (slot (this, &Location_UI::down));

	start_grab_tape.clicked.connect
		(slot (this, &Location_UI::set_start_from_tape));
	end_grab_tape.clicked.connect
		(slot (this, &Location_UI::set_end_from_tape));

	auto_punch_indicator->clicked.connect
		(slot (this, &Location_UI::auto_punch_toggle));
	auto_loop_indicator->clicked.connect
		(slot (this, &Location_UI::auto_loop_toggle));

	start_button.button_release_event.connect
		(slot (this, &Location_UI::locate_start));

	end_button.button_release_event.connect
		(slot (this, &Location_UI::locate_end));

	set_next_button.button_release_event.connect
		(slot (this, &Location_UI::set_next_from_tape));

	start_adjustment.value_changed.connect
		(slot (this, &Location_UI::set_start));

	end_adjustment.value_changed.connect
		(slot (this, &Location_UI::set_end));

	start_backward.button_press_event.connect
		(bind (slot (this, &Location_UI::fake_button_press), 1));
	start_backward.button_release_event.connect
		(slot (start_spinner, &Gtkmmext::AutoSpin::stop_spinning));

	start_forward.button_press_event.connect
		(bind (slot (this, &Location_UI::fake_button_press), 2));
	start_forward.button_release_event.connect
		(slot (start_spinner, &Gtkmmext::AutoSpin::stop_spinning));

	end_backward.button_press_event.connect
		(bind (slot (this, &Location_UI::fake_button_press), 3));
	end_backward.button_release_event.connect
		(slot (end_spinner, &Gtkmmext::AutoSpin::stop_spinning));

	end_forward.button_press_event.connect
		(bind (slot (this, &Location_UI::fake_button_press), 4));
	end_forward.button_release_event.connect
		(slot (end_spinner, &Gtkmmext::AutoSpin::stop_spinning));

	set_usize_to_display_given_text 
		(start_display, "88:88:88:88", 10, 10);
	set_usize_to_display_given_text 
		(end_display, "88:88:88:88", 10, 10);

	set_name ("TransportBase");
	put (hpacker, 0, 0);
}

gint
Location_UI::fake_button_press (GdkEventButton *ev, int which)

{
	gfloat step_inc;
	gfloat page_inc;

	if (tape == 0) {
		return TRUE;
	}

	/* This is here to fake out the AutoSpinners. We pretend
	   that any press on the "back" button was done using
	   mouse button 1, which the AutoSpinners will use
	   as a decrement. Likewise, any press on the "fwd" button
	   will appear to be mouse button 3, which will increment.

	   Before we do that though, we use button state (Ctrl, Meta,
	   Alt, Shift etc.) to change the spin rate.
	*/

	/* all values in frames:

	   - control gives sample level and 1msec accuracy
	   - shift   gives 1msec and 100th sec accuracy
           - meta/alt gives crude control - 10secs, 2 minutes
	   - no mods gives 1 second and 1 minute
	*/

   	if (ev->state & GDK_CONTROL_MASK) {
		step_inc = 1;
		page_inc = tape->frame_rate() / 1000;
	} else if (ev->state & GDK_SHIFT_MASK) {
		step_inc = tape->frame_rate() / 1000;
		page_inc = tape->frame_rate() / 100;
	} else if (ev->state & GDK_MOD1_MASK || ev->state & GDK_MOD2_MASK) {
		step_inc = tape->frame_rate() * 10;
		page_inc = tape->frame_rate() * 120;
	} else {
		step_inc = tape->frame_rate();
		page_inc = tape->frame_rate() * 60;
	}
	
	/* remove control state, because that does something we don't
	   really want.
	*/

	ev->state &= ~GDK_CONTROL_MASK;

	switch (which) {
	case 1:
		ev->button = 1;
		start_adjustment.set_step_increment (step_inc);
		start_adjustment.set_page_increment (page_inc);
		return start_spinner.button_press (ev);
		break;
	case 2:
		ev->button = 3;
		start_adjustment.set_step_increment (step_inc);
		start_adjustment.set_page_increment (page_inc);
		return start_spinner.button_press (ev);
		break;
	case 3:
		ev->button = 1;
		end_adjustment.set_step_increment (step_inc);
		end_adjustment.set_page_increment (page_inc);
		return end_spinner.button_press (ev);
		break;
	case 4:
		ev->button = 3;
		end_adjustment.set_step_increment (step_inc);
		end_adjustment.set_page_increment (page_inc);
		return end_spinner.button_press (ev);
		break;

	default:
		fatal << "Programming error in Location_UI::fake_button_press"
		      << endmsg;
		/*NOTREACHED*/
		return TRUE;
	}
}

void
Location_UI::up ()

{
	if (tape) {
		tape->locations()->up();
	}
}

void
Location_UI::down ()

{
	if (tape) {
		tape->locations()->down ();
	}
}

void
Location_UI::set_location_name ()

{
	if (yes_its_me) return;
	location->set_name (name_entry.get_text());
}

void
Location_UI::set_start_from_tape ()

{
	start_adjustment.set_value (tape->current_frame ());
}

void
Location_UI::set_end_from_tape ()

{
	end_adjustment.set_value (tape->current_frame());
}

gint
Location_UI::set_next_from_tape (GdkEventButton *ev)

{
	/* Move to next location */

	up ();

	/* set the start time */

	location->set_start (tape->current_frame ());

	return TRUE;
}

void
Location_UI::set_start ()

{
	if (yes_its_me) return;
	size_t start = (size_t) start_adjustment.get_value();
	location->set_start (start);
}

void
Location_UI::set_end ()

{
	if (yes_its_me) return;
	size_t end = (size_t) end_adjustment.get_value();
	location->set_end (end);
}

void
Location_UI::touch ()

{
	/* This is only called when a transport state change
	   causes a non-GUI thread to reset the last stop
	   or record locations. So, we pick up the new values,
	   preventing recursive hell in the process, and redisplay
	   those values right here.
	*/

	yes_its_me = true;
	start_adjustment.set_value (location->start());
	end_adjustment.set_value (location->end());
	yes_its_me = false;
	map_start (location);
	map_end (location);
}

void
Location_UI::map_start (Location *l)

{
	char buf[128];

	if (location != l) {
		return;
	}

	if (!ARDOUR_UI::instance()->caller_is_gui_thread()) {
		ARDOUR_UI::instance()->touch_display (this);
		return;
	}

	sample_to_smpte_clock (location->start(),
			       buf,
			       mcd->frame_rate(),
			       Config->smpte_frames_per_second);
	start_display.set_text (buf);
}

void
Location_UI::map_end (Location* l)

{
	char buf[128];

	if (location != l) {
		return;
	}

	if (!ARDOUR_UI::instance()->caller_is_gui_thread()) {
		ARDOUR_UI::instance()->touch_display (this);
		return;
	}

	sample_to_smpte_clock (location->end(),
			       buf,
			       mcd->frame_rate(),
			       Config->smpte_frames_per_second);
	end_display.set_text (buf);
}

void
Location_UI::map_auto_punch (Location* l)

{
	if (location->is_auto_punch ()) {
		auto_punch_indicator->set_active (true);
	} else {
		auto_punch_indicator->set_active (false);
	}
}

void
Location_UI::map_auto_loop (Location* l)

{
	if (location->is_auto_loop ()) {
		auto_loop_indicator->set_active (true);
	} else {
		auto_loop_indicator->set_active (false);
	}
}

void
Location_UI::map_name (Location *l)

{
	if (location != l) {
		return;
	}

	yes_its_me = true;
	name_entry.set_text (location->name());
	yes_its_me = false;
}

void
Location_UI::location_change (Location *l)

{
	name_connection.disconnect ();
	start_connection.disconnect ();
	end_connection.disconnect ();

	location = l;

	name_connection = location->name_changed.connect
		(slot (this, &Location_UI::map_name));
	start_connection = location->start_changed.connect
		(slot (this, &Location_UI::map_start));
	end_connection = location->end_changed.connect
		(slot (this, &Location_UI::map_end));

	/* ensure that the adjustments have the right value */

	start_adjustment.gtkobj()->value = location->start();
	end_adjustment.gtkobj()->value = location->end();

	map_start (location);
	map_end (location);
	map_name (location);
	map_auto_punch (location);
	map_auto_loop (location);

	/* and jump to it if we're stationary */

	if (tape->transport_state() == Tape::Stopped) {
		ARDOUR_UI::instance()->transport_locate (location->start());
	}
}

void
Location_UI::connect_to_tape (ARDOUR::Tape *t)

{
	tape = t;

	tape->going_away.connect (slot (this, &Location_UI::disconnect_from_tape));

	/* stop the adjustment from taking us beyond the end
	   of the tape. Seems like a good idea, for now.
	*/

	end_adjustment.gtkobj()->upper = tape->frames ();
	start_adjustment.gtkobj()->upper = tape->frames ();

	tape->locations()->changed.connect 
		(slot (this, &Location_UI::location_change));

	location_change (tape->locations()->current());
}

void
Location_UI::disconnect_from_tape (AudioThing *t)

{
	tape = 0;
	auto_loop_indicator->set_active (false);
	auto_punch_indicator->set_active (false);
	start_display.set_text ("");
	end_display.set_text ("");
	name_entry.set_text ("");
}

void
Location_UI::auto_punch_toggle ()

{
	if (tape) {
		if (auto_punch_indicator->get_active()) {
			tape->set_auto_punch_location (location);
		} else {
			tape->remove_auto_punch_location (location);
		}
	}
}

void
Location_UI::auto_loop_toggle ()

{
	if (tape) {
		if (auto_loop_indicator->get_active()) {
			tape->set_auto_loop_location (location);
		} else {
			tape->remove_auto_loop_location (location);
		}
	}
}

gint
Location_UI::locate_start (GdkEventButton *ev)

{
	if (tape) {
		ARDOUR_UI::instance()->transport_locate 
			(location->start());
	}
	return TRUE;
}

gint
Location_UI::locate_end (GdkEventButton *ev)

{
	if (tape) {
		ARDOUR_UI::instance()->transport_locate 
			(location->end());
	}
	return TRUE;
}
