/*
    Copyright (C) 1999 Paul Barton-Davis 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: track_ui.cc,v 1.40 2000/10/28 13:27:38 pbd Exp $
*/

#include <stdio.h> /* sprintf */
#include <math.h>
#include <time.h>
#include <sys/stat.h>
#include <algo.h>

#include <pbd/pathscanner.h>
#include <gtkmmext/utils.h>
#include <gtkmmext/selector.h>

#include "track_ui.h"
#include "ardour_ui.h"
#include "alsa_device.h"
#include "track.h"
#include "utils.h"

using namespace ARDOUR;
using namespace Gtk;
using namespace Gtkmmext;

Tooltips *Track_UI::track_tooltips = 0;
Gtk::Window *Track_UI::take_window;
Gtk::Entry *Track_UI::take_name_display;
Gtk::Entry *Track_UI::track_name_display;
static Track_UI *lister_track_ui;

static void
take_printer (char buf[32], Gtk::Adjustment &adj, void *ignored)

{
	gfloat val = adj.get_value();

	if (val < 100) {
		sprintf (buf, "%d", (int) val);
	} else {
		sprintf (buf, "%.2f", val/1000.0);
	} 
}

static void
channel_printer (char buf[32], Gtk::Adjustment &adj, void *ignored)

{
	gfloat val = adj.get_value();

	if (val < 0) {
		buf[0] = '-';
		buf[1] = '\0';
	} else {
		sprintf (buf, "%d", (int) val + 1);
	}
}

Track_UI::Track_UI (ALSA::MultiChannelDevice *mcdp, size_t tn)
	: mcd (mcdp),
	  tape (0),
	  track (0),
	  channel_num (tn),
	  rec_enable (ARDOUR_UI::instance()->rec_enable_button_pix(), 
		     "RecPunchButton", false),
	  monitor_toggle (ARDOUR_UI::instance()->monitor_button_pix(), 
		     "MonitorToggleButton", false),
	  meter (200, 8, GTK_LEVEL_HOLD_FADE, 5, 4, Config->meter_hold),
	  take_button (ARDOUR_UI::instance()->small_push_button_pix(),
		       "TakeButton", false),
	  inchan_adjustment (-1, -1, mcdp->nchannels() - 1, 1, 8, 0),
	  inchan (&inchan_adjustment, "ChannelSelector"),
	  outchan_adjustment (-1, -1, mcdp->nchannels() - 1, 1, 8, 0),
	  outchan (&outchan_adjustment, "ChannelSelector")
	  
{
	char buf[64];

	if (track_tooltips == 0) {
		
		/* static member initialization */

		create_tooltips ();
		create_menus_and_dialogs ();
	}

	sprintf (buf, "%d", channel_num+1);
	label.set_text (buf);
	label.set_name ("ChannelMeterLabel");

	meter_hbox.pack_start (meter, true, false);

	/* initialize button states.
	 */

	rec_enable.set_sensitive (false);
	monitor_toggle.set_active (false);
	monitor_toggle_pending = false;
	
	inchan.set_print_func (channel_printer, 0);
	inchan_hbox.pack_start (inchan);
	inchan_frame.add (inchan_hbox);
	inchan_frame.set_shadow_type (GTK_SHADOW_IN);
	inchan_frame.set_name ("BaseFrame");

	track_tooltips->set_tip (inchan, "input channel");
	inchan_adjustment.value_changed.connect
		(slot(this,&Track_UI::inchan_adjusted));
	
	outchan.set_print_func (channel_printer, 0);
	outchan_hbox.pack_start (outchan);
	outchan_frame.add (outchan_hbox);
	outchan_frame.set_shadow_type (GTK_SHADOW_IN);
	outchan_frame.set_name ("BaseFrame");

	track_tooltips->set_tip (outchan, "output channel");
	outchan_adjustment.value_changed.connect
		(slot(this,&Track_UI::outchan_adjusted));

	take_hbox.pack_start (take_button);

	track_tooltips->set_tip (take_button, "take #");
	take_button.clicked.connect (slot(this,&Track_UI::take_clicked));

	over_long_label.set_text ("0");
	over_long_label.set_name ("OverMeterLabel");
	over_long_button.add (over_long_label);
	over_long_button.set_name ("OverMeter");
	over_long_frame.add (over_long_button);
	over_long_frame.set_shadow_type (GTK_SHADOW_IN);
	over_long_frame.set_name ("BaseFrame");
	over_long_hbox.pack_start (over_long_frame, false, false);

	sprintf (buf, "# of %d-sample overs", Config->over_length_long);
	track_tooltips->set_tip (over_long_button, buf);

	set_usize_to_display_given_text (over_long_button, "888", 2, 2);

	over_short_label.set_text ("0");
	over_short_label.set_name ("OverMeterLabel");
	over_short_button.add (over_short_label);
	over_short_button.set_name ("OverMeter");
	over_short_frame.add (over_short_button);
	over_short_frame.set_shadow_type (GTK_SHADOW_IN);
	over_short_frame.set_name ("BaseFrame");
	over_short_hbox.pack_start (over_short_frame, false, false);

	sprintf (buf, "# of %d-sample overs", Config->over_length_short);
	track_tooltips->set_tip (over_short_button, buf);

	set_usize_to_display_given_text (over_short_button, "888", 2, 2);

	above_meter_vbox.set_spacing (5);
	below_meter_vbox.set_spacing (5);

	above_meter_vbox.pack_start (over_long_hbox, false, false);
	above_meter_vbox.pack_start (over_short_hbox, false, false);

	/* start with the led "turned off" */

	input_ok_led.black();
	input_ok_led_box.pack_start (input_ok_led, true, false);

	track_tooltips->set_tip (input_ok_led, "input state");

	rec_enable_hbox.pack_start (rec_enable, true, false);
	monitor_hbox.pack_start (monitor_toggle, true, false);

	track_tooltips->set_tip (rec_enable, "record enable");
	track_tooltips->set_tip (monitor_toggle, "monitor toggle");

	if (mcd->has_clock_lock_reporting()) {
		below_meter_vbox.pack_start (input_ok_led_box, false, false);
	}
	below_meter_vbox.pack_start (inchan_frame, false, false);
	below_meter_vbox.pack_start (outchan_frame, false, false);
	below_meter_vbox.pack_start (rec_enable_hbox);
	below_meter_vbox.pack_start (monitor_hbox);
	below_meter_vbox.pack_start (take_hbox);
	below_meter_vbox.pack_start (label);

	rec_enable.button_press_event.connect 
		(slot(this,&Track_UI::toggle_rec));
	monitor_toggle.button_release_event.connect
		(slot(this,&Track_UI::toggle_monitor));
	over_short_button.button_release_event.connect
		(slot(this,&Track_UI::gui_clear_overs));
	over_long_button.button_release_event.connect
		(slot(this,&Track_UI::gui_clear_overs));

	mcd->input_monitor_change.connect
		(slot (*this, &Track_UI::input_monitor_change));

	last_over_short = 0;
	last_over_long = 0;

	meter_clear_pending = false;
	over_clear_pending = false;
}

void
Track_UI::record_enable_blink (bool onoff)

{
	if (track == 0) {
		return;
	}

	if (track->record_enabled()) {

		Tape::RecordState rs;
		
		rs = tape->record_status ();

		switch (rs) {
		case Tape::Disabled:
		case Tape::Enabled:
			if (onoff) {
				rec_enable.set_state (GTK_STATE_SELECTED);
			} else {
				rec_enable.set_state (GTK_STATE_NORMAL);
			}
			break;

		case Tape::Recording:
			if (rec_enable.get_state() != GTK_STATE_SELECTED) {
				rec_enable.set_state (GTK_STATE_SELECTED);
			}
			break;
		}

	} else {
		if (rec_enable.get_state() != GTK_STATE_NORMAL) {
			rec_enable.set_state (GTK_STATE_NORMAL);
		}
	}
}

void
Track_UI::touch ()

{
	/* This is called by the GUI thread in response to a
	   TouchDisplay request involving this object. Its used
	   when some other thread needed us to update our
	   display in a way that was not based on the meter
	   timeout.
	*/

	if (tape->transport_state() == Tape::Stopped) {
		take_button.set_sensitive (true);
	}  else {
		take_button.set_sensitive (false);
	}
}

void
Track_UI::update ()

{
	char buf[32];
	size_t over_short = 0;
	size_t over_long = 0;
	int chn = -1;
	float power = 0;

	if (track)  {

		if (track->monitoring_input ()) {
			
			if (tape->transport_state() == Tape::Roll) {
				if (!tape->get_auto_input()) {
					
					/* Rolling, but no auto-input,
					   so use MCD metering 
					*/

					if ((chn = track->input_channel()) >= 0) {
						power = mcd->peak_input_power (chn);
					} else {
						/* its zero - input channel not set */
					}

				} else {

					if (tape->record_status() >= Tape::Enabled) {

						/* Rolling, Recording,
						   with auto-input:
						   use mcd metering
						*/

						if ((chn = track->input_channel()) >= 0) {
							power = mcd->peak_input_power (chn);
						} else {
							/* its zero - input channel not set */
						}
						
					} else {

						/* Rolling, not Recording with auto-input,
						   so use tape/track metering.
						*/

						power = track->peak_power ();
					}
				}

			} else {

				/* Not rolling - use MCD metering */
				
				if ((chn = track->input_channel()) >= 0) {
					power = mcd->peak_input_power (chn);
				} else {
					/* its zero - input channel not set */
				}
			}

		} else {
			/* Not metering input */

			power = track->peak_power ();
		}

		if (monitor_toggle_pending) {
			if (!track->record_enabled()) {
				monitor_toggle.set_active (track->monitoring_input());
			}
		}

	} else {

		if (channel_num < mcd->nchannels()) {
			if (mcd->monitor_in (channel_num)) {
				power = mcd->peak_input_power (channel_num);
				chn = channel_num;
			} 

			if (monitor_toggle_pending) {
				monitor_toggle.set_active (mcd->monitor_in (channel_num));
			}
		}
	}

	meter.set (power);

	if (track && chn >= 0) {
		over_short = mcd->over_count 
			(chn, ALSA::MultiChannelDevice::ShortOver);
		over_long = mcd->over_count 
			(chn, ALSA::MultiChannelDevice::LongOver);

		if (over_short != last_over_short) {
			sprintf (buf, "%d", over_short);
			over_short_label.set_text (buf);
			last_over_short = over_short;
		}
		
		if (over_long != last_over_long) {
			sprintf (buf, "%d", over_long);
			over_long_label.set_text (buf);
			last_over_long = over_long;
		}
	}

	if (meter_clear_pending) {
		meter_clear_pending = false;
		meter.clear ();
	}

	if (over_clear_pending) {
		over_clear_pending = false;
		mcd->reset_overs (channel_num);
		over_long_label.set_text ("0");
		over_short_label.set_text ("0");
		last_over_short = 0;
		last_over_long = 0;
	}

	monitor_toggle_pending = false;
}

gint
Track_UI::gui_clear_overs (GdkEventButton *ev)

{
	clear_overs ();
	return FALSE;
}

void
Track_UI::clear_meter ()

{
	meter_clear_pending = true;
}

void
Track_UI::clear_overs ()

{
	over_clear_pending = true;
}

void
Track_UI::connect_to_tape (Tape *t)

{
	size_t inchan;
	size_t outchan;
	
	tape = t;
	track = t->track (channel_num);

	t->going_away.connect (slot (this, &Track_UI::disconnect_from_tape));
	t->TransportStateChange.connect (slot (this, &Track_UI::map_transport_state));

	if (track) {
		track->take_changed.connect (slot (this, &Track_UI::take_or_tag_changed));
		track->tag_changed.connect (slot (this, &Track_UI::take_or_tag_changed));

		/* We have to set the initial take ID/tooltip because
		   the Track object would have loaded it (and sent
		   `take_changed') already. Subsequent changes are
		   picked up via the signal shown above.  
		*/

		take_or_tag_changed ();

		inchan = track->input_channel();
		inchan_adjustment.set_value (inchan);
		
		outchan = track->output_channel();
		outchan_adjustment.set_value (outchan);

		allow_record_enable ();

		/* tapes come up with input monitoring disabled
		   by default.
		*/

		monitor_toggle.set_active (false);
		rec_enable.set_state (GTK_STATE_NORMAL);

	} else {
		inchan_adjustment.set_value (channel_num);
		outchan_adjustment.set_value (-1);
	}
	
	clear_meter ();
}

void
Track_UI::disconnect_from_tape (AudioThing *t)

{
	tape = 0;
	track = 0;
	inchan_adjustment.set_value (-1);
	outchan_adjustment.set_value (-1);
	disallow_record_enable ();
}

void
Track_UI::map_transport_state ()

{
	ARDOUR_UI::instance()->touch_display (this);
}

void
Track_UI::take_or_tag_changed ()

{
	char buf[64];

	sprintf (buf, "%s: take %u", 
		 track->tagname().c_str(), 
		 track->current_take());
	track_tooltips->set_tip (take_button, buf);
}

gint
Track_UI::toggle_rec (GdkEventButton *ev)

{
	if (track == 0) {
		goto no_regular_handler;
	}

	tape->set_track_record_enable (channel_num, !track->record_enabled());

	if (track->record_enabled()) {
		rec_enable.set_state (GTK_STATE_SELECTED);
	} else {
		rec_enable.set_state (GTK_STATE_NORMAL);
	}

  no_regular_handler:

	/* This stops the button press from having any effect on the
	   button itself.
	*/

	gtk_signal_emit_stop_by_name (GTK_OBJECT(rec_enable.gtkobj()),
				      "button-press-event");
	return TRUE;
}

gint
Track_UI::toggle_monitor (GdkEventButton *ev)

{
	bool yn = monitor_toggle.get_active();

	if (track) {
		track->set_monitor_input (yn);
	} else {
		mcd->request_monitor_input (channel_num, yn);
	}

	clear_meter ();
	return FALSE;
}

void
Track_UI::input_monitor_change (size_t chn, bool yn)

{
	if (chn == channel_num) {
		monitor_toggle_pending = true;
	}
}

void
Track_UI::inchan_adjusted ()

{
	int chn;

	clear_meter ();

	if (track == 0) {
		return;
	}

	if ((chn = (int) inchan_adjustment.get_value()) !=
	    track->input_channel ()) {
		tape->set_track_input_channel (channel_num, chn);
	}

	return;
}

void
Track_UI::outchan_adjusted ()

{
	int chn;

	clear_meter ();

	if (track == 0) {
		return;
	}

	if ((chn = (int) outchan_adjustment.get_value ()) !=
	    track->output_channel ()) {
		tape->set_track_output_channel (channel_num, chn);
	}
	return;
}

void
Track_UI::take_clicked ()

{
	if (track != 0) {
		char buf[64];

		sprintf (buf, "Track %u (take: %u)", channel_num + 1, track->current_take());
		take_window->set_title (buf);
		track_name_display->set_text (track->tagname());
//		take_name_display->set_text (???); NOT IMPLEMENTED

		lister_track_ui = this;
		take_window->show_all ();
	}
}

void
Track_UI::allow_record_enable ()

{
	rec_enable.set_sensitive (true);
}

void
Track_UI::disallow_record_enable ()

{
	rec_enable.set_sensitive (false);
}

void
Track_UI::create_tooltips ()

{
	track_tooltips = new Gtk::Tooltips;
	track_tooltips->enable();
}

bool
Track_UI::take_filter (const string &str)

{
	size_t trk, take;

	return  ((sscanf (str.c_str(), "%u.%u", &trk, &take) == 2) && 
		 (strspn (str.c_str(), "0123456789.") == str.length()) &&
		 (trk == lister_track_ui->channel_num + 1));
}

void 
Track_UI::take_lister (Gtk::CList &list, void *arg)
{
	vector<string *> *matched_files;
	vector<string *>::iterator i;
	PathScanner scanner;
	int row;
	time_t now;

	matched_files = scanner (lister_track_ui->tape->path(), take_filter, false, true);

	if (matched_files == 0) {
		return;
	}

	if (matched_files->size() == 0) {
		delete matched_files;
		return;
	}

	time (&now);

	for (row = 0, i = matched_files->begin(); 
	     i != matched_files->end(); 
	     row++, i++) {

		const gchar *rowdata[3];
		string &fullpath = *(*i);
		int start, end;
		size_t take;
		size_t elapsed;

		/* now get the time-since-modified */
		
		struct stat statbuf;
		char elapsed_str[64];

		if (stat (fullpath.c_str(), &statbuf)) {
			error << "File \""
			      << fullpath
			      << "\" disappeared after we found it!"
			      << endmsg;
			continue;
		}

		elapsed = now - statbuf.st_mtime;
		elapsed_time_to_str (elapsed_str, elapsed);

		start = fullpath.find_last_of ('.') + 1;
		end = fullpath.length();
		
		string takestr = fullpath.substr (start, (end-start));
		rowdata[0] = takestr.c_str();
		rowdata[1] = ""; /* TAKE NAME - not implemented */
		rowdata[2] = elapsed_str;
		list.insert_row (row, rowdata);
		
		/* Now set the data pointer to be the take number */
		
		sscanf (rowdata[0], "%u", &take);

		list.set_row_data (row, (void *) take);
	}
}

void
Track_UI::create_menus_and_dialogs ()

{
	Gtkmmext::PixmapButton *new_button = new Gtkmmext::PixmapButton
		(ARDOUR_UI::instance()->small_push_button_pix(),
		 "TakeMenuButton");
	Gtkmmext::PixmapButton *erase_button = new Gtkmmext::PixmapButton
		(ARDOUR_UI::instance()->small_push_button_pix(),
		 "TakeMenuButton");
	Gtkmmext::PixmapButton *cancel_button = new Gtkmmext::PixmapButton
		(ARDOUR_UI::instance()->small_push_button_pix(),
		 "TakeMenuButton");

	Gtkmmext::Selector *take_selector;
	Gtk::Frame         *take_frame;
	Gtk::VBox          *take_vpacker;

	Gtk::Label *new_label = new Gtk::Label ("Use a new take");
	Gtk::Label *erase_label = new Gtk::Label ("Erase the current take");
	Gtk::Label *cancel_label = new Gtk::Label ("Cancel");

	new_label->set_name ("TakeLabel");
	erase_label->set_name ("TakeLabel");
	cancel_label->set_name ("TakeLabel");

	Gtk::HBox *new_hbox = new Gtk::HBox;
	new_hbox->pack_start (*new_button, false, false, 10);
	new_hbox->pack_start (*new_label, false, false, 10);

	Gtk::HBox *erase_hbox = new Gtk::HBox;
	erase_hbox->pack_start (*erase_button, false, false, 10);
	erase_hbox->pack_start (*erase_label, false, false, 10);

	Gtk::HBox *cancel_hbox = new Gtk::HBox;
	cancel_hbox->pack_start (*cancel_button, false, false, 10);
	cancel_hbox->pack_start (*cancel_label, false, false, 10);

	const gchar *selector_titles[] = {
		"#",
		"Name",
		"Time Since Modified"
	};

	take_selector = new Gtkmmext::Selector (take_lister, 0, 3,
						false, true,
						selector_titles);
	take_selector->set_usize (100, 100);

	take_window = new Gtk::Window (GTK_WINDOW_DIALOG);
	take_window->set_name ("TakeWindow");
	take_window->set_position (GTK_WIN_POS_MOUSE);
	take_window->set_modal (true);

	take_frame = new Gtk::Frame;
	take_vpacker = new Gtk::VBox;

	Gtk::HBox *selector_hbox = new Gtk::HBox;
	selector_hbox->pack_start (*take_selector, true, true, 10);
	take_selector->clist().set_name ("TakeSelector");

	take_window->add (*take_frame);
	take_window->set_title ("ardour: take operations");
	take_frame->add (*take_vpacker);
	take_frame->set_shadow_type (GTK_SHADOW_IN);
	take_frame->set_name ("BaseFrame");

	track_name_display = new Gtk::Entry;

	Gtk::Frame *track_name_frame = new Gtk::Frame;
	Gtk::HBox  *track_name_hbox = new Gtk::HBox;
	Gtk::Label *track_name_label = new Gtk::Label ("TRACK TAG");

	track_name_display->set_name ("TakeSelector");
	track_name_label->set_name ("TakeTagLabel");
	track_name_frame->add (*track_name_display);
	track_name_hbox->pack_start (*track_name_label, true, false, 10);
	track_name_hbox->pack_start (*track_name_frame, true, true, 20);
	track_name_frame->set_shadow_type (GTK_SHADOW_IN);
	track_name_frame->set_name ("BaseFrame");

	take_vpacker->set_spacing (5);
	take_vpacker->pack_start (*track_name_hbox, false, false, 10);

	take_name_display = new Gtk::Entry;

	Gtk::Frame *take_name_frame = new Gtk::Frame;
	Gtk::HBox  *take_name_hbox = new Gtk::HBox;
	Gtk::Label *take_name_label = new Gtk::Label ("TAKE TAG");

	take_name_display->set_name ("TakeSelector");
	take_name_label->set_name ("TakeTagLabel");
	take_name_frame->add (*take_name_display);
	take_name_hbox->pack_start (*take_name_label, true, false, 10);
	take_name_hbox->pack_start (*take_name_frame, true, true, 10);
	take_name_frame->set_shadow_type (GTK_SHADOW_IN);
	take_name_frame->set_name ("BaseFrame");

	take_vpacker->pack_start (*take_name_hbox, false, false, 10);

	Gtk::HSeparator *sep1 = new Gtk::HSeparator;
	Gtk::HBox *sepbox = new Gtk::HBox;

	sep1->set_name ("PrimaryLogo");
	sepbox->pack_start (*sep1, true, true, 10);
	take_vpacker->pack_start (*sepbox, false, false, 10);

	take_vpacker->pack_start (*selector_hbox, false, false);
	take_vpacker->pack_start (*new_hbox, false, false);
	take_vpacker->pack_start (*erase_hbox, false, false);
	take_vpacker->pack_start (*cancel_hbox, false, false, 10);

	new_button->clicked.connect (slot (this, &Track_UI::new_take));
	erase_button->clicked.connect (slot (this, &Track_UI::erase_take));
	cancel_button->clicked.connect (slot (this, &Track_UI::take_cancelled));
	take_name_display->changed.connect (slot (this, &Track_UI::take_tag_display_changed));
	track_name_display->changed.connect (slot (this, &Track_UI::track_tag_display_changed));
	take_selector->selection_made.connect (slot (Track_UI::take_selected));
}

void
Track_UI::take_selected (Gtkmmext::Selector *selector,
			 SelectionResult *result)

{
	lister_track_ui->select_take ((size_t) result->data);
}

void
Track_UI::select_take (size_t take)

{
	take_window->hide_all ();

	if (track) {
		track->use_take (take);
	}
}

void
Track_UI::track_tag_display_changed ()

{
	if (track) {
		string new_tag = track_name_display->get_text();

		if (new_tag.length()) {
			track->set_tagname (new_tag);
		}
	}
}

void
Track_UI::take_tag_display_changed ()

{
	if (track) {
		string new_tag = take_name_display->get_text();

		if (new_tag.length()) {
// TAKE TAG NOT IMPLEMENTED
//			track->set_tagname (new_tag);
		}
	}
}

void
Track_UI::take_cancelled ()

{
	take_window->hide_all ();
}

void
Track_UI::new_take ()
{
	take_window->hide_all ();

	if (track) {
		track->use_new_take ();
	}
}

void
Track_UI::erase_take ()
{
	take_window->hide_all  ();
	/* get confirmation before doing this */
}
